import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';
import { OrbitControls } from './jsm/controls/OrbitControls.js';
import { RGBELoader } from './jsm/loaders/RGBELoader.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.outputEncoding = THREE.sRGBEncoding;
        renderer.toneMapping = THREE.ACESFilmicToneMapping;
        renderer.toneMappingExposure = 0.75;
        renderer.setPixelRatio(window.devicePixelRatio);
        domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        //scene.background = new THREE.Color(0xffffff);
        this._scene = scene;

        this._setupFont();
        //this._setupModel();
        //this._setupInteractions();
        //this._setupCamera();
        //this._setupLight();
        //this._setupGUI();
        //this._setupControls();
        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        //this.resize();
        //this.render()
    }

    _setupFont() {
        const loader = new THREE.FontLoader();
        loader.load("fonts/gentilis_regular.typeface.json", (font) => {
            this._setupModel(font);
            this._setupCamera();
            this._setupLight();
            this._setupGUI();
            this._setupControls();

            this.resize();
            this.render();
        });
    }

    _setupInteractions() {
        //.
    }

    _setupModel(font) {
        const pmremGenerator = new THREE.PMREMGenerator(this._renderer);
		pmremGenerator.compileEquirectangularShader();
        
        new RGBELoader()
            .setDataType(THREE.UnsignedByteType)
            .setPath("textures/equirectangular/")
            .load("pedestrian_overpass_1k.hdr", (hdrEquirect) => {
                const hdrCubeRenderTarget = pmremGenerator.fromEquirectangular(hdrEquirect);
                hdrEquirect.dispose();
                pmremGenerator.dispose();
                //this._scene.background = hdrCubeRenderTarget.texture;

                const cubeWidth = 400;
                const numberOfSpheresPerSide = 5;
                const sphereRadius = (cubeWidth / numberOfSpheresPerSide) * 0.8 * 0.5;
                const stepSize = 1.0 / numberOfSpheresPerSide;
                
                const geometry = new THREE.SphereGeometry(sphereRadius, 32, 16);

				let imgTexture = new THREE.TextureLoader().load( "textures/planets/moon_1024.jpg" );
				imgTexture.wrapS = imgTexture.wrapT = THREE.RepeatWrapping;
				imgTexture.anisotropy = 16;

                let index = 0;
                for(let alpha=0; alpha<=1.0; alpha+=stepSize) {
                    for(let beta=0; beta<=1.0; beta+=stepSize) {
                        for(let gamma=0; gamma<=1.0; gamma+=stepSize) {
                            const diffuseColor = new THREE.Color().setHSL(alpha, 0.5, gamma*0.5+0.1);
                            
                            const material = new THREE.MeshStandardMaterial({
                                map: imgTexture,
                                bumpMap: imgTexture,
                                bumpScale: 1,
                                color: diffuseColor,
                                metalness: beta,
                                roughness: 1.0 - alpha,
                                envMap: index % 2 === 0 ? null : hdrCubeRenderTarget.texture
                            });

                            index ++;

                            const mesh = new THREE.Mesh(geometry, material);

                            mesh.position.x = alpha * 400 - 200;
                            mesh.position.y = beta * 400 - 200;
                            mesh.position.z = gamma * 400 - 200;

                            this._scene.add(mesh);
                        }
                    }
                    index++;
                }
            }
        );

        this._addLabel(font, "+roughness", new THREE.Vector3(-350, 0, 0));
        this._addLabel(font, "-roughness", new THREE.Vector3(350, 0, 0));
        
        this._addLabel(font, "-metalness", new THREE.Vector3(0, -300, 0));
        this._addLabel(font, "+metalness", new THREE.Vector3(0, 300, 0));
        
        this._addLabel(font, "-diffuse", new THREE.Vector3(0, 0, -300));
        this._addLabel(font, "+diffuse", new THREE.Vector3(0, 0, 300));

        const particleLight = new THREE.Mesh(
            new THREE.SphereGeometry(4, 8, 8), 
            new THREE.MeshBasicMaterial({color:0xffffff})
        );

        this._scene.add(particleLight);

        this._particleLight = particleLight;
    }

    _addLabel(font, name, location) {
        const textGeo = new THREE.TextGeometry(name, {
            font: font,
            size: 20,
            height: 1,
            curveSegments: 1
        });

        const textMaterial = new THREE.MeshBasicMaterial({ color: 0xffffff });
        const textMesh = new THREE.Mesh(textGeo, textMaterial);
        textMesh.position.copy(location);

        this._scene.add(textMesh);
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }

    _setupCamera() {    
       const camera = new THREE.PerspectiveCamera(40, window.innerWidth/window.innerHeight, 1, 2500);
       camera.position.set(0.0, 400, 400*3.5);
       this._camera = camera;
    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this._renderer.domElement);
        controls.minDistance = 200;
        controls.maxDistance = 2000;
    }

    _setupLight() {
        this._scene.add(new THREE.AmbientLight(0x222222));
        
        const directionalLight = new THREE.DirectionalLight(0xffffff, 1);
        directionalLight.position.set(1, 1, 1).normalize();
        this._scene.add(directionalLight);

        const pointLight = new THREE.PointLight(0xffffff, 2, 800);
        this._particleLight.add(pointLight);
    }

    update() {
        const timer = Date.now() * 0.00025;
        //this._camera.lookAt(this._scene.position);

        this._particleLight.position.x = Math.sin(timer*7)*300;
        this._particleLight.position.y = Math.cos(timer*5)*400;
        this._particleLight.position.z = Math.cos(timer*3)*300;

        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    new App();
}